/*
 * mbox_get_message.c
 * Get (parts of) a specific message from a mbox
 * libpostal project
 * libpost_raw mbox code
 * $Id: mbox_get_message.c,v 1.30 2003/07/16 11:50:49 fullermd Exp $
 */

#include <sys/file.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <post_raw.h>
#include <post_raw_p.h>
#include <post_raw_mbox_p.h>
#include <post_raw_extra_p.h>

/* 16k.  This can be tuned, and should be made tunable. */
#define READ_CHUNK	16384


/*
 * TARGETS FOR OPTIMIZATION:
 * - In GET_HEADER case, when we have Content-Length:, we can
 *   short-circuit a lot of stuff.
 * - If msg->*len is already defined, we can use it as a hint and save
 *   some time and over- or repetitive-read()'s.
 * - Make the initial READ_CHUNK smaller to grab just a few k of the
 *   headers, and then check for Content-Length:.  In the case where it's
 *   there, this can save us the extra read() size.  This MAY turn out to
 *   be a wild goose chase, though, because the time to read() 128k from
 *   any modern disk is practically nil compared to inherent latencies.
 */

int
mbox_get_message(POSTAL_CONN *p_conn, POSTAL_MSG *msg, int get_type)
{
	char *rawmsg=NULL, *rawmtmp, *rawloc;
	char *buf;
	char *hdrs;
	char *fromloc;
	int fromdist=0;
	int notfrom;
	int gotmsg=0;
	int firsttime=1;
	ssize_t readlen, readsz;
	size_t bufavail, bufsz;
	/*
	 * XXX These should maybe be size_t's or something.
	 * However, we compare (conlen<0) later, and size_t is unsigned, so
	 * don't muff with this unless you fix that.
	 */
	long conlen=0;
	long hdrlen=0;
	long morelen;

	/* Get to where we 'think' we're starting */
	lseek(p_conn->spec.mbox->desc, msg->loc.offset, SEEK_SET);

	/*
	 * XXX NOTE!
	 * We assume that nothing has changed since these
	 * messages were pulled out of the mailbox!
	 * Fixme later.
	 */

	/* We're going to use READ_CHUNK as our base unit */
	rawmsg = malloc(READ_CHUNK);
	if(rawmsg==NULL)
		POSTAL_ERRSET_RI(POSTAL_E_NOMEM, NULL);
	rawloc = rawmsg;
	fromloc = rawmsg;
	bufavail = READ_CHUNK;
	bufsz = READ_CHUNK;

	while(gotmsg==0)
	{
		/* Read in a block */
		if( (readlen = read(p_conn->spec.mbox->desc, rawloc, (bufavail-1)))
				> 0)
		{
			rawloc[readlen] = '\0';
			bufavail -= readlen;
		}
		else if (readlen==0 && firsttime==1)
		{
			/* EOF on the first time */
			free(rawmsg);
			POSTAL_ERRSET_RI(POSTAL_E_NOMSG, NULL);
		}
		else if (readlen<0)
		{
			/* Read error, Bad Thing (tm).  Bail */
			free(rawmsg);
			POSTAL_ERRSET_RI(POSTAL_E_FSOP, NULL);
		}

		/* Check for Content-Length: if it's the first loop */
		if(firsttime==1)
		{
			/*
			 * Sigh.  mbox's are a PITA.
			 * We need to check for Content-Length: if it's
			 * there and use it, otherwise we may not be able to
			 * correctly subdivide the message.
			 *
			 * Have I mentioned lately that mbox sucks?
			 */
			while( !(buf=strstr(rawmsg, "\n\n")) )
			{
				/*
				 * What, no HEADERS?  With 128k BLOCKSIZE?!
				 * Try a bit more...
				 */
				rawmtmp = realloc(rawmsg, (bufsz + READ_CHUNK));
				if(rawmtmp==NULL)
				{
					free(rawmsg);
					POSTAL_ERRSET_RI(POSTAL_E_NOMEM, NULL);
				}
				rawmsg = rawmtmp;
				bufsz += READ_CHUNK;
				bufavail += READ_CHUNK;
				fromloc = rawmsg;
				rawloc = rawmsg + (bufsz - bufavail);
				readlen = read(p_conn->spec.mbox->desc, rawloc,
						(bufavail-1));
				if(readlen==0)
				{
					free(rawmsg);
					POSTAL_ERRSET_RI(POSTAL_E_DFORM, NULL);
				}
				else if(readlen==-1)
				{
					free(rawmsg);
					POSTAL_ERRSET_RI(POSTAL_E_FSOP, NULL);
				}
				rawloc[readlen] = '\0';
				bufavail -= readlen;
			}
	
			/*
			 * We have a full header block.
			 * This lets us include the \n at the end, but not the \n
			 * after it (before the body).
			 */
			buf++;
			hdrlen = buf-rawmsg;
			hdrs = malloc(hdrlen+1);
			if(hdrs==NULL)
			{
				free(rawmsg);
				POSTAL_ERRSET_RI(POSTAL_E_NOMEM, NULL);
			}
			strncpy(hdrs, rawmsg, hdrlen);
			hdrs[hdrlen]='\0';

			/* Check for Content-Length: */
			buf=postal_get_header(hdrs, "Content-Length:");
			if( (buf==NULL) && (postal_errno()!=POSTAL_E_NOERR) )
			{
				free(hdrs);
				free(rawmsg);
				POSTAL_ERRRET_RI; /* Preserve p_errno */
			}
			if(buf!=NULL)
			{
				conlen = atol(buf);
				free(buf);

				/*
				 * Now, we know we have the Content-Length.
				 * Let's start processing on it.
				 * Content-Length is the length of the body of the
				 * message, in bytes, not including the blank
				 * line between messages in a mbox.
				 *
				 * See if we read the entire message.
				 * Note: this does NOT include the blank line
				 * at the end.
				 */
				/* -1 for '\0' */
				readsz = bufsz - bufavail - 1;
				if( readsz < (hdrlen + conlen + 1) )
				{
					/* Bah, we don't have the whole thing */
					morelen = (hdrlen + conlen + 1) - readsz;
					/* +1 for '\0' */
					rawmtmp = realloc( rawmsg, (bufsz + morelen + 1) );
					if(rawmtmp==NULL)
					{
						free(hdrs);
						free(rawmsg);
						POSTAL_ERRSET_RI(POSTAL_E_NOMEM, NULL);
					}
					rawmsg = rawmtmp;
					bufsz += morelen+1;
					bufavail += morelen+1;
					fromloc=rawmsg;
					rawloc = rawmsg + (bufsz - bufavail);
					readlen = read(p_conn->spec.mbox->desc, rawloc,
							(bufavail-1));
					if(readlen==0)
					{
						free(hdrs);
						free(rawmsg);
						POSTAL_ERRSET_RI(POSTAL_E_DFORM, NULL);
					}
					else if(readlen==-1)
					{
						free(hdrs);
						free(rawmsg);
						POSTAL_ERRSET_RI(POSTAL_E_FSOP, NULL);
					}
					rawloc[readlen]='\0';
					bufavail -= readlen;

					/*
					 * Sanity check: bufavail should be 0.
					 * If it's not, then read() didn't do us right.
					 */
					if(bufavail!=0)
					{
						free(hdrs);
						free(rawmsg);
						POSTAL_ERRSET_RI(POSTAL_E_FSOP,
								"Didn't read() enough");
					}

					/*
					 * We should have the whole message now.
					 * Operative word being 'should'.  We 'should'
					 * add more sanity checks.
					 */
					gotmsg=1;
				}
				else
				{
					/*
					 * We already have the whole thing, and probably
					 * then some.  Extra-long read()'s will be
					 * backed-down later, so just delimit the body, and
					 * move on down the road.
					 */
					rawmsg[(hdrlen+conlen+1)] = '\0';
					gotmsg=1;
				}
			} /* Have Content-Length:? */
		} /* First-time check for Content-Length: */

		firsttime = 0;

		/*
		 * If we enter here, we don't have Content-Length.
		 * We'll have to fake it, and search for a From_ line.
		 *
		 * See if we have the beginning of the next message.
		 * If we do, cut off right before it, adjust sizing,
		 * and all that jazz.
		 * If we don't, get read to read() more.
		 */
		if((readlen==0) && (gotmsg==0))
		{
			/* At EOF */
			/* Do this better.  strlen() is EEEEVIL */
			size_t tmplen = strlen(rawmsg);
			gotmsg=1;
			rawmsg[(tmplen-1)]='\0';
			conlen = tmplen-(hdrlen+1);
		}
		else if( (buf=strstr(fromloc, "\n\nFrom ")) && (gotmsg==0) )
		{
			/*
			 * We've either got a full message with the
			 * beginning of the next, or we're at EOF.
			 * Either way, we've got a full-enough message.
			 */

			buf+=2;
			/* This is to handle looping back around */
			fromloc=buf;
			fromdist = fromloc-rawmsg;

			notfrom=0;
			/*
			 * XXX Special case hack.
			 * From_ 'should' be, say, at least 80 chars.
			 * Oooh, this sucks.  We should bring over the macros from
			 * mbox_get_message_all(), or find a more general way to do
			 * this.
			 */
			if(strlen(buf)<80)
			{
				goto endloop;
			}

			/* Verify it */
			notfrom=mbox_validate_from(buf);

			if(notfrom==0)
			{
				/*
				 * We've got a full message with one starting after.
				 * Record the extra length and lseek() back to the
				 * beginning of the next message.
				 */
				buf[-1] = '\0';
				conlen=(buf-1)-(rawmsg+hdrlen+1);
				gotmsg=1;
			}
		}

endloop:
		/* Read more if we don't have it yet */
		if(gotmsg==0)
		{
			/* Alloc more */
			rawmtmp = realloc(rawmsg, (bufsz + READ_CHUNK));
			if(rawmtmp==NULL)
			{
				free(hdrs);
				free(rawmsg);
				POSTAL_ERRSET_RI(POSTAL_E_NOMEM, NULL);
			}
			rawmsg = rawmtmp;
			bufsz += READ_CHUNK;
			bufavail += READ_CHUNK;
			rawloc = rawmsg + (bufsz - bufavail);
			fromloc = rawmsg + fromdist;
		}
	}

	/*
	 * Once here, we've got a full message ready to parse out.
	 * Set some vars and move to the right spot.
	 */
	lseek(p_conn->spec.mbox->desc,
			(msg->loc.offset + (strlen(rawmsg)+1)), SEEK_SET);

	/* Shovel the data into the struct */
	if( (get_type==GET_HEADER) || (get_type == GET_BODY) )
	{
		msg->header = hdrs;
		msg->hdrlen = hdrlen;
		msg->stat_loc = mbox_get_status(msg);
		if(msg->stat_loc==-1)
		{
			free(hdrs);
			free(rawmsg);
			msg->header=NULL;
			msg->hdrlen=0;
			POSTAL_ERRRET_RI; /* Preserve p_errno */
		}
	}
	else
		free(hdrs);

	if( (get_type==GET_BODY) )
	{
		/* Because we may have re-used 'buf' above */
		buf = rawmsg + hdrlen + 1;

		/*
		 * We've got 1 extra in both cases because of the extra \n in
		 * mboxen.
		 */
		conlen--;
		/* If there's no body, this could end up as -1. */
		if(conlen<0)
			conlen=0;
		msg->body = malloc(conlen+1);
		if(msg->body==NULL)
		{
			free(hdrs);
			free(rawmsg);
			msg->header=NULL;
			msg->hdrlen=0;
			POSTAL_ERRSET_RI(POSTAL_E_NOMEM, NULL);
		}
		msg->bodylen = conlen;
		buf = rawmsg+hdrlen+1;
		strncpy(msg->body, buf, conlen);
		msg->body[conlen] = '\0';
	}

	/* We're all filled in and it worked fine.  Clean up and return */
	free(rawmsg);
	return(0);
}
